#!/usr/bin/env python3
"""
URL Scanner
Web security scanner for URL analysis and vulnerability detection.
Parameters:
    - url: Target URL to scan
    - depth: Crawling depth for linked pages
    - timeout: Request timeout in seconds
Requires: requests, beautifulsoup4, urllib3 packages
"""
import requests
import argparse
import logging
from datetime import datetime
import json
import threading
import queue
from bs4 import BeautifulSoup
from urllib.parse import urljoin, urlparse
import re
import ssl
import socket
import dns.resolver
import whois
from concurrent.futures import ThreadPoolExecutor

class URLScanner:
    def __init__(self, url, depth=1, timeout=10, threads=10):
        self.base_url = url
        self.depth = depth
        self.timeout = timeout
        self.threads = threads
        self.visited_urls = set()
        self.results_queue = queue.Queue()
        self.setup_logging()
        self.setup_session()

    def setup_logging(self):
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s'
        )
        self.logger = logging.getLogger(__name__)

    def setup_session(self):
        self.session = requests.Session()
        self.session.headers.update({
            'User-Agent': 'Mozilla/5.0 URLScanner/1.0'
        })

    def is_same_domain(self, url):
        """Check if URL belongs to the same domain as base_url"""
        return urlparse(url).netloc == urlparse(self.base_url).netloc

    def get_domain_info(self, domain):
        """Gather domain information"""
        info = {'domain': domain}
        
        try:
            # DNS records
            for record_type in ['A', 'MX', 'NS', 'TXT']:
                try:
                    answers = dns.resolver.resolve(domain, record_type)
                    info[f'{record_type}_records'] = [str(r) for r in answers]
                except:
                    info[f'{record_type}_records'] = []

            # WHOIS information
            try:
                w = whois.whois(domain)
                info['whois'] = {
                    'registrar': w.registrar,
                    'creation_date': str(w.creation_date),
                    'expiration_date': str(w.expiration_date),
                    'name_servers': w.name_servers
                }
            except:
                info['whois'] = None

        except Exception as e:
            self.logger.error(f"Error getting domain info: {str(e)}")

        return info

    def check_ssl(self, domain):
        """Check SSL certificate information"""
        try:
            context = ssl.create_default_context()
            with socket.create_connection((domain, 443)) as sock:
                with context.wrap_socket(sock, server_hostname=domain) as ssock:
                    cert = ssock.getpeercert()
                    return {
                        'issuer': dict(x[0] for x in cert['issuer']),
                        'subject': dict(x[0] for x in cert['subject']),
                        'version': cert['version'],
                        'serialNumber': cert['serialNumber'],
                        'notBefore': cert['notBefore'],
                        'notAfter': cert['notAfter']
                    }
        except Exception as e:
            self.logger.error(f"Error checking SSL: {str(e)}")
            return None

    def check_security_headers(self, headers):
        """Check for security-related HTTP headers"""
        security_headers = {
            'Strict-Transport-Security': 'Missing HSTS',
            'Content-Security-Policy': 'Missing CSP',
            'X-Frame-Options': 'Missing X-Frame-Options',
            'X-Content-Type-Options': 'Missing X-Content-Type-Options',
            'X-XSS-Protection': 'Missing XSS Protection',
            'Referrer-Policy': 'Missing Referrer Policy'
        }
        
        results = {}
        for header, message in security_headers.items():
            if header in headers:
                results[header] = headers[header]
            else:
                results[header] = message
        
        return results

    def find_vulnerabilities(self, url, content, headers):
        """Check for common vulnerabilities"""
        vulnerabilities = []
        
        # Check for exposed email addresses
        emails = re.findall(r'[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}',
                          content)
        if emails:
            vulnerabilities.append({
                'type': 'Information Disclosure',
                'details': f'Exposed email addresses: {", ".join(emails)}'
            })

        # Check for common debugging files
        debug_files = ['/debug', '/phpinfo.php', '/test.php', '/server-status']
        for path in debug_files:
            try:
                r = self.session.head(urljoin(url, path),
                                   timeout=self.timeout,
                                   allow_redirects=False)
                if r.status_code == 200:
                    vulnerabilities.append({
                        'type': 'Information Disclosure',
                        'details': f'Potential debug file exposed: {path}'
                    })
            except:
                continue

        # Check for directory listing
        if 'Index of /' in content:
            vulnerabilities.append({
                'type': 'Information Disclosure',
                'details': 'Directory listing enabled'
            })

        # Check for form submissions without CSRF tokens
        soup = BeautifulSoup(content, 'html.parser')
        forms = soup.find_all('form')
        for form in forms:
            if not form.find('input', {'type': 'hidden'}):
                vulnerabilities.append({
                    'type': 'CSRF',
                    'details': f'Form without CSRF token: {form.get("action", "")}'
                })

        return vulnerabilities

    def scan_url(self, url, depth=0):
        """Scan a single URL for security issues"""
        if url in self.visited_urls or depth > self.depth:
            return

        self.visited_urls.add(url)
        
        try:
            response = self.session.get(url, timeout=self.timeout)
            content = response.text
            
            result = {
                'url': url,
                'status_code': response.status_code,
                'headers': dict(response.headers),
                'security_headers': self.check_security_headers(response.headers),
                'vulnerabilities': self.find_vulnerabilities(
                    url, content, response.headers
                ),
                'links': []
            }

            # Parse links for crawling
            if depth < self.depth:
                soup = BeautifulSoup(content, 'html.parser')
                for link in soup.find_all('a'):
                    href = link.get('href')
                    if href:
                        absolute_url = urljoin(url, href)
                        if self.is_same_domain(absolute_url):
                            result['links'].append(absolute_url)
                            if absolute_url not in self.visited_urls:
                                self.scan_url(absolute_url, depth + 1)

            self.results_queue.put(result)
            
        except Exception as e:
            self.logger.error(f"Error scanning {url}: {str(e)}")

    def save_results(self, output_file):
        results = []
        while not self.results_queue.empty():
            results.append(self.results_queue.get())

        domain = urlparse(self.base_url).netloc
        scan_results = {
            'timestamp': datetime.now().isoformat(),
            'base_url': self.base_url,
            'domain_info': self.get_domain_info(domain),
            'ssl_info': self.check_ssl(domain),
            'pages_scanned': len(results),
            'findings': results
        }

        try:
            with open(output_file, 'w') as f:
                json.dump(scan_results, f, indent=4)
            self.logger.info(f"Results saved to {output_file}")
        except Exception as e:
            self.logger.error(f"Error saving results: {str(e)}")

    def print_summary(self):
        results = []
        while not self.results_queue.empty():
            results.append(self.results_queue.get())

        self.logger.info("\nScan Summary:")
        self.logger.info(f"Total URLs scanned: {len(results)}")
        
        # Count vulnerabilities by type
        vuln_count = {}
        for result in results:
            for vuln in result['vulnerabilities']:
                vuln_type = vuln['type']
                vuln_count[vuln_type] = vuln_count.get(vuln_type, 0) + 1
        
        if vuln_count:
            self.logger.info("\nVulnerabilities Found:")
            for vuln_type, count in vuln_count.items():
                self.logger.info(f"{vuln_type}: {count}")
        else:
            self.logger.info("No vulnerabilities found")

        # Security headers analysis
        missing_headers = set()
        for result in results:
            for header, value in result['security_headers'].items():
                if value.startswith('Missing'):
                    missing_headers.add(header)
        
        if missing_headers:
            self.logger.info("\nMissing Security Headers:")
            for header in missing_headers:
                self.logger.info(f"- {header}")

    def start_scan(self, output_file=None):
        self.logger.info(f"Starting scan of {self.base_url}")
        self.scan_url(self.base_url)
        
        if output_file:
            self.save_results(output_file)
        
        self.print_summary()

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="URL Scanner")
    parser.add_argument("url", help="Target URL to scan")
    parser.add_argument("--depth", type=int, default=1,
                      help="Crawling depth for linked pages")
    parser.add_argument("--timeout", type=int, default=10,
                      help="Request timeout in seconds")
    parser.add_argument("--threads", type=int, default=10,
                      help="Number of concurrent threads")
    parser.add_argument("--output", help="Output file for scan results")
    args = parser.parse_args()

    scanner = URLScanner(args.url, args.depth, args.timeout, args.threads)
    scanner.start_scan(args.output)
